﻿using System;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml;
using System.Xml.Linq;
using PureCloud.ChatBot.Service.Models;
using jabber.client;

namespace PureCloud.ChatBot.Service
{
	public class JabberService
	{
		#region Private Members
		private readonly string _user;
		private readonly string _server;
		private readonly string _password;
		private readonly string _networkHost;
		private readonly bool _ignoreInvalidCertificate;
		private readonly ICommandHandler _commandHandler;
		#endregion Private Members

		#region Constructors
		public JabberService(
			string user,
			string server,
			string password,
			string networkHost,
			bool ignoreInvalidCertificate,
			ICommandHandler commandHandler)
		{
			_user = user;
			_server = server;
			_password = password;
			_networkHost = networkHost;
			_ignoreInvalidCertificate = ignoreInvalidCertificate;
			_commandHandler = commandHandler;
		}
		#endregion Constructors

		#region Public Methods
		public void Start()
		{
			Initialize();
		}

		public void Stop()
		{
		}
		#endregion Public Methods

		#region Event Handlers
		private void OnAuthenticate(object sender)
		{
        }

		private void OnWriteText(object sender, string text)
		{
			/*
			if (text == " ")
			{
				return;
			}
			*/
		}

		private void OnReadText(object sender, string text)
		{
			/*
			if (text == " ")
			{
				return;
			}
			*/
		}
		
		private void OnError(object sender, Exception ex)
		{
		    Console.WriteLine(ex);
		}

		private bool OnInvalidCertificate(object sender, System.Security.Cryptography.X509Certificates.X509Certificate certificate, System.Security.Cryptography.X509Certificates.X509Chain chain, System.Net.Security.SslPolicyErrors sslPolicyErrors)
		{
			//_trace.Error("Invalid certificate ({}):\n{}", sslPolicyErrors.ToString(), certificate.ToString(true));
			return true;
		}

		private void OnMessage(object sender, jabber.protocol.client.Message msg)
		{
			try
			{
				JabberClient j = (JabberClient)sender;
					
				string roomJid;
				if (IsRoomInvite(msg, out roomJid))
				{
					SendJoinRoom(j, msg, roomJid);
					return;
				}

				if (string.IsNullOrEmpty(msg.Body))
				{
					return;
				}

				string inputCmd = "";

				Match m = Regex.Match(msg.Body.ToLower(), @"\/\S+");
				if (m.Success)
				{
					inputCmd = m.Value;
				}

				if (inputCmd == "/kick"  && msg.From.ToString().Contains("@conference"))
				{
					SendMessageWithActiveState(j, msg, "Goodbye!");
					SendLeaveRoom(j, msg);
					return;
				}

				SendComposingState(j, msg);

				CommandResult result = _commandHandler.Handle(msg, inputCmd);

				if (String.IsNullOrEmpty(result.Output))
				{
					SendActiveState(j, msg);
					return;
				}

				SendMessageWithActiveState(j, msg, result.Output);
			}
			catch (Exception ex)
			{
				Console.WriteLine(ex);

				var j = sender as JabberClient;
				if (j != null)
				{
					SendActiveState(j, msg);
				}
			}
	}
		#endregion Event Handlers

		#region Private Methods
		private void Initialize()
		{
			JabberClient j = new JabberClient
				{
					User = _user,
					Server = _server,
					Password = _password,
					NetworkHost = _networkHost,
					AutoPresence = true,
					AutoRoster = true,
					AutoStartTLS = true,
					AutoStartCompression = true,
					AutoReconnect = 1,
				};

			j.OnError += OnError;
			j.OnAuthenticate += OnAuthenticate;
			j.OnReadText += OnReadText;
			j.OnWriteText += OnWriteText;
			j.OnMessage += OnMessage;

			if (_ignoreInvalidCertificate)
			{
				j.OnInvalidCertificate += OnInvalidCertificate;
			}

			var cm = new jabber.connection.CapsManager
				{
					Stream = j,
					Node = "http://www.inin.com/IT/ChatService",
				};
			cm.AddFeature("http://jabber.org/protocol/muc");
			cm.AddFeature("http://jabber.org/protocol/disco#info");

			j.Connect();
	    }

		private void SendComposingState(JabberClient j, jabber.protocol.client.Message msg)
		{
			XmlDocument doc = new XmlDocument();

			var messageType = msg.From.ToString().Contains("@conference")
				? jabber.protocol.client.MessageType.groupchat
				: jabber.protocol.client.MessageType.chat;

			string to = messageType == jabber.protocol.client.MessageType.groupchat
				? msg.From.Bare
				: msg.From.ToString();

			jabber.protocol.client.Message message = new jabber.protocol.client.Message(doc)
				{
					To = to,
					From = msg.To,
					Type = messageType,
				};

			AppendChatState(doc, message, "composing");

			j.Write(message);
		}

		private void SendActiveState(JabberClient j, jabber.protocol.client.Message msg)
		{
			var messageType = msg.From.ToString().Contains("@conference")
				? jabber.protocol.client.MessageType.groupchat
				: jabber.protocol.client.MessageType.chat;

			string to = messageType == jabber.protocol.client.MessageType.groupchat
				? msg.From.Bare
				: msg.From.ToString();

			jabber.protocol.client.Message message = new jabber.protocol.client.Message(j.Document)
				{
					To = to,
					From = msg.To,
					Type = messageType,
				};

			AppendChatState(j.Document, message, "active");

			j.Write(message);
		}

		private void SendMessageWithActiveState(JabberClient j, jabber.protocol.client.Message msg, string body)
		{
			var messageType = msg.From.ToString().Contains("@conference") 
				? jabber.protocol.client.MessageType.groupchat 
				: jabber.protocol.client.MessageType.chat;

			string to = messageType == jabber.protocol.client.MessageType.groupchat
				? msg.From.Bare
				: msg.From.ToString();

			jabber.protocol.client.Message message = new jabber.protocol.client.Message(j.Document)
				{
					To = to,
					From = msg.To,
					Type = messageType,
					Body = body,
				};

			AppendChatState(j.Document, message, "active");

			j.Write(message);
		}

		private bool IsRoomInvite(jabber.protocol.client.Message msg, out string roomJid)
		{
			if(msg.Name == "message")
			{
				XElement element = XElement.Parse(msg.OuterXml);

				var conferenceElement = element.Elements().FirstOrDefault(x => x.Name.LocalName.Equals("x"));
				if (conferenceElement != null)
				{
					if (conferenceElement.GetDefaultNamespace() == "jabber:x:conference")
					{
						roomJid = conferenceElement.Attribute("jid").Value;

						return true;
					}
				}
			}

			roomJid = string.Empty;

			return false;
		}

		private void SendJoinRoom(JabberClient j, jabber.protocol.client.Message msg, string roomJid)
		{
			string nick = msg.To.Bare.Replace('@', '#');

			jabber.protocol.iq.RoomPresence roomPresence = new jabber.protocol.iq.RoomPresence(j.Document, roomJid + "/" + nick)
				{
					From = msg.To
				};

			j.Write(roomPresence);
		}

		private void SendLeaveRoom(JabberClient j, jabber.protocol.client.Message msg)
		{
			string nick = msg.To.Bare.Replace('@', '#');

			jabber.protocol.client.Presence presence = new jabber.protocol.client.Presence(j.Document)
				{
					From = msg.To,
					To = msg.From.Bare + "/" + nick,
					Type = jabber.protocol.client.PresenceType.unavailable
				};

			j.Write(presence);
		}

		private void AppendChatState(XmlDocument doc, XmlElement element, string state)
		{
			XmlElement chatState = (XmlElement) element.AppendChild(doc.CreateElement(state));
			chatState.SetAttribute("xmlns", "http://jabber.org/protocol/chatstates");
		}

		private string FormatXml(string xml)
		{
			try
			{
				XDocument doc = XDocument.Parse(xml);
				return doc.ToString();
			}
			catch (Exception)
			{
				return xml;
			}
		}
		#endregion Private Methods
	}
}
